/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotSensorMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD
#include <atomic>
#include <thread>
#include <queue>

// Simox
#include <SimoxUtility/meta/enum/adapt_enum.h>

// BaseClass
#include <ArmarXCore/core/Component.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/RobotUnitComponentPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/RobotStateComponentPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/PlatformUnitComponentPlugin.h>
#include <RobotAPI/libraries/RobotUnitDataStreamingReceiver/RobotUnitDataStreamingReceiver.h>
#include <RobotAPI/libraries/armem/server/ComponentPlugin.h>

// ArmarX
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>


namespace armarx
{
    /**
     * @defgroup Component-RobotSensorMemory RobotSensorMemory
     * @ingroup RobotAPI-Components
     * A description of the component RobotSensorMemory.
     *
     * @class RobotSensorMemory
     * @ingroup Component-RobotSensorMemory
     * @brief Brief description of class RobotSensorMemory.
     *
     * Detailed description of class RobotSensorMemory.
     */
    class RobotSensorMemory :
        virtual public armarx::Component,
        virtual public armem::server::ComponentPluginUser,
        virtual public RobotUnitComponentPluginUser,
        virtual public RobotStateComponentPluginUser

    {
    public:
        RobotSensorMemory();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

    protected:
        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;

        /// @see PropertyUser::createPropertyDefinitions()
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

    private:
        // RobotUnit Streaming
        void setupRobotUnitSegment();

        void convertRobotUnitStreamingDataToAron();
        void storeConvertedRobotUnitDataInMemory();

        void startRobotUnitStream();
        void stopRobotUnitStream();

        // RobotStateComponent Poll
        void setupRobotStateComponentSegment();

        void convertRobotStateComponentDataToAron();
        void storeConvertedRobotStateComponentDataInMemory();

        void startRobotStateComponentPoll();
        void stopRobotStateComponentPoll();

    private:
        std::string memoryName = "RobotSensorMemory";
        bool addCoreSegmentOnUsage = false;

        mutable std::recursive_mutex startStopMutex;

        // Memory IDs
        std::string robotUnitCoreSegmentName = "Proprioception";
        std::string robotUnitProviderSegmentName = "RobotUnit"; // get robot name?
        armem::data::MemoryID robotUnitProviderID;

        std::string robotStateComponentCoreSegmentName = "Localization";
        std::string robotStateComponentProviderSegmentName = "RobotStateComponent"; // get robot name?
        armem::data::MemoryID robotStateComponentProviderID;

        // RobotUnit stuff
        RobotUnitDataStreaming::DataStreamingDescription keys;
        std::vector<RobotUnitDataStreaming::DataEntry> keysList;
        RobotUnitDataStreaming::Config cfg;
        RobotUnitDataStreamingReceiverPtr handler;

        struct RobotUnitData
        {
            struct RobotUnitDataGroup
            {
                long timestamp;
                std::string name;
                aron::datanavigator::DictNavigatorPtr data;
            };

            std::map<std::string, RobotUnitDataGroup> groups;
        };

        // params
        static constexpr int ROBOT_UNIT_MAXIMUM_FREQUENCY = 100;
        int robotUnitPollFrequency = 50;
        std::string robotUnitSensorPrefix = "sens.*";
        unsigned int robotUnitMemoryBatchSize = 50;

        // queue
        std::queue<RobotUnitData> robotUnitDataQueue;
        mutable std::mutex robotUnitDataMutex;

        // running tasks
        armarx::PeriodicTask<RobotSensorMemory>::pointer_type robotUnitConversionTask;
        armarx::PeriodicTask<RobotSensorMemory>::pointer_type robotUnitStoringTask;

        // RobotStateComponent stuff
        struct RobotStateComponentData
        {
            long timestamp;
            aron::datanavigator::DictNavigatorPtr data;
        };

        // params
        static constexpr int ROBOT_STATE_COMPONENT_MAXIMUM_FREQUENCY = 100;
        int robotStateComponentPollFrequency = 50;
        unsigned int robotStateComponentMemoryBatchSize = 50;

        // queue
        std::queue<RobotStateComponentData> robotStateComponentDataQueue;
        mutable std::mutex robotStateComponentDataMutex;

        // running tasks
        armarx::PeriodicTask<RobotSensorMemory>::pointer_type robotStateComponentConversionTask;
        armarx::PeriodicTask<RobotSensorMemory>::pointer_type robotStateComponentStoringTask;
    };
}
