#include "MemoryFileSystemStorage.h"

#include <RobotAPI/libraries/aron/core/Debug.h>
#include <RobotAPI/libraries/armem/core/error.h>


namespace armarx::armem::io
{
    FileSystemMemoryManager::FileSystemMemoryManager(const std::filesystem::path& rootPath, bool createFolder) :
        FileSystemMemoryManager(std::make_shared<io::NlohmannJSONDiskWriter>(rootPath, createFolder),
                                std::make_shared<io::NlohmannJSONDiskReader>(rootPath, createFolder))
    {
    }

    FileSystemMemoryManager::FileSystemMemoryManager(const io::DiskWriterPtr& w, const io::DiskReaderPtr& r) :
        writer(w),
        reader(r)
    {
    }

    void FileSystemMemoryManager::writeOnDisk(const Memory& m)
    {
        io::DiskWriterReturnInformation info = writer->writeOnDisk(m);
        merge(info.storedElements);
    }

    void FileSystemMemoryManager::writeOnDisk(const MemoryID& id, const CoreSegment& s)
    {
        io::DiskWriterReturnInformation info = writer->writeOnDisk(id, s);
        merge(info.storedElements);
    }

    void FileSystemMemoryManager::writeOnDisk(const MemoryID& id, const ProviderSegment& s)
    {
        io::DiskWriterReturnInformation info = writer->writeOnDisk(id, s);
        merge(info.storedElements);
    }

    void FileSystemMemoryManager::writeOnDisk(const MemoryID& id, const Entity& s)
    {
        io::DiskWriterReturnInformation info = writer->writeOnDisk(id, s);
        merge(info.storedElements);
    }

    void FileSystemMemoryManager::writeOnDisk(const MemoryID& id, const EntitySnapshot& s)
    {
        io::DiskWriterReturnInformation info = writer->writeOnDisk(id, s);
        merge(info.storedElements);
    }

    Memory FileSystemMemoryManager::readMemoryFromDisk(const std::string& n) const
    {
        MemoryID id;
        id.memoryName = n;
        return readMemoryFromDisk(id);
    }

    Memory FileSystemMemoryManager::readMemoryFromDisk(const MemoryID& _id) const
    {
        if (!_id.hasMemoryName())
        {
            throw error::InvalidMemoryID(_id, "MemoryName is missing.");
        }
        Memory ret(_id.memoryName);
        if (internalMemory.hasMemory(_id))
        {
            MemoryID id = _id.getMemoryID();
            for (const auto& [coreKey, _] : internalMemory.getDiskMemory(_id))
            {
                id.coreSegmentName = coreKey;
                CoreSegment s = readCoreSegmentFromDisk(id);
                ret.addCoreSegment(s);
            }
        }
        return ret;
    }

    CoreSegment FileSystemMemoryManager::readCoreSegmentFromDisk(const MemoryID& _id) const
    {
        if (!_id.hasCoreSegmentName())
        {
            throw error::InvalidMemoryID(_id, "CoreSegmentName is missing.");
        }

        CoreSegment ret(_id.coreSegmentName);
        if (internalMemory.hasCoreSegment(_id))
        {
            MemoryID id = _id.getCoreSegmentID();
            for (const auto& [providerKey, _] : internalMemory.getDiskCoreSegment(_id))
            {
                id.providerSegmentName = providerKey;
                ProviderSegment s = readProviderSegmentFromDisk(id);
                ret.addProviderSegment(s);
            }
        }
        return ret;
    }

    ProviderSegment FileSystemMemoryManager::readProviderSegmentFromDisk(const MemoryID& _id) const
    {
        if (!_id.hasProviderSegmentName())
        {
            throw error::InvalidMemoryID(_id, "ProviderSegmentName is missing.");
        }

        ProviderSegment ret(_id.providerSegmentName);
        if (internalMemory.hasProviderSegment(_id))
        {
            MemoryID id = _id.getProviderSegmentID();
            for (const auto& [entityKey, _] : internalMemory.getDiskProviderSegment(_id))
            {
                id.entityName = entityKey;
                Entity s = readEntityFromDisk(id);
                ret.addEntity(s);
            }
        }
        return ret;
    }

    Entity FileSystemMemoryManager::readEntityFromDisk(const MemoryID& _id) const
    {
        if (!_id.hasEntityName())
        {
            throw error::InvalidMemoryID(_id, "EntityName is missing.");
        }

        Entity ret(_id.entityName);
        if (internalMemory.hasEntity(_id))
        {
            MemoryID id = _id.getEntityID();
            for (const auto& [ts, _] : internalMemory.getDiskEntity(_id))
            {
                id.timestamp = ts;
                EntitySnapshot s = readEntitySnapshotFromDisk(id);
                ret.addSnapshot(s);
            }
        }
        return ret;
    }

    EntitySnapshot FileSystemMemoryManager::readEntitySnapshotFromDisk(const MemoryID& _id) const
    {
        if (!_id.hasTimestamp())
        {
            throw error::InvalidMemoryID(_id, "Timestamp is missing.");
        }

        EntitySnapshot ret(_id.timestamp);
        if (internalMemory.hasEntitySnapshot(_id))
        {
            MemoryID id = _id.getEntitySnapshotID();
            for (unsigned int i = 0; i < internalMemory.getDiskEntitySnapshot(_id).size(); ++i)
            {
                id.instanceIndex = static_cast<int>(i);
                EntityInstance s = readEntityInstanceFromDisk(id);
                ret.addInstance(s);
            }
        }
        return ret;
    }

    EntityInstance FileSystemMemoryManager::readEntityInstanceFromDisk(const MemoryID& _id) const
    {
        if (!_id.hasInstanceIndex())
        {
            throw error::InvalidMemoryID(_id, "InstanceIndex is missing.");
        }

        EntityInstance ret(_id.instanceIndex);
        if (internalMemory.hasEntityInstance(_id))
        {
            aron::typenavigator::ObjectNavigatorPtr typeInformation = readClosestTypeInformation(_id);
            const DiskEntityInstance& path = internalMemory.getDiskEntityInstance(_id);
            return reader->readSingleInstanceFromDisk(path, typeInformation);
        }
        return ret;
    }

    aron::typenavigator::ObjectNavigatorPtr FileSystemMemoryManager::readClosestTypeInformation(const MemoryID& _id) const
    {
        if (_id.hasInstanceIndex())
        {
            aron::typenavigator::ObjectNavigatorPtr t = reader->readSingleTypeInformationFromDisk(_id);
            if (t != nullptr)
            {
                return t;
            }
        }
        if (_id.hasTimestamp())
        {
            MemoryID id = _id.getEntitySnapshotID();
            aron::typenavigator::ObjectNavigatorPtr t = reader->readSingleTypeInformationFromDisk(id);
            if (t != nullptr)
            {
                return t;
            }
        }
        if (_id.hasEntityName())
        {
            MemoryID id = _id.getEntityID();
            aron::typenavigator::ObjectNavigatorPtr t = reader->readSingleTypeInformationFromDisk(id);
            if (t != nullptr)
            {
                return t;
            }
        }
        if (_id.hasProviderSegmentName())
        {
            MemoryID id = _id.getProviderSegmentID();
            aron::typenavigator::ObjectNavigatorPtr t = reader->readSingleTypeInformationFromDisk(id);
            if (t != nullptr)
            {
                return t;
            }
        }
        if (_id.hasCoreSegmentName())
        {
            MemoryID id = _id.getCoreSegmentID();
            aron::typenavigator::ObjectNavigatorPtr t = reader->readSingleTypeInformationFromDisk(id);
            if (t != nullptr)
            {
                return t;
            }
        }
        if (_id.hasMemoryName())
        {
            MemoryID id = _id.getMemoryID();
            aron::typenavigator::ObjectNavigatorPtr t = reader->readSingleTypeInformationFromDisk(id);
            if (t != nullptr)
            {
                return t;
            }
        }

        return nullptr;
    }

    void FileSystemMemoryManager::update()
    {
        FileSystemLookupMemory up = reader->readMemoryStructureFromDisk();
        merge(up);
    }

    std::string FileSystemMemoryManager::toString()
    {
        std::stringstream ss;
        std::string tabs = "";
        for (const auto& [memoryName, memory] : internalMemory.memoryLookupTable)
        {
            ss << memoryName << ": " << std::endl;
            std::string tabs = "\t";
            for (const auto& [coreKey, coreSegment] : memory)
            {
                ss << tabs << coreKey << ": " << std::endl;
                std::string tabs = "\t\t";
                for (const auto& [providerKey, providerSegment] : coreSegment)
                {
                    ss << tabs << providerKey << ": " << std::endl;
                    std::string tabs = "\t\t\t";
                    for (const auto& [entityKey, entity] : providerSegment)
                    {
                        ss << tabs << entityKey << ": " << std::endl;
                        std::string tabs = "\t\t\t\t";
                        for (const auto& [entityHistoryTimestamp, entitySnapshot] : entity)
                        {
                            ss << tabs << entityHistoryTimestamp.toMicroSeconds() << ":" << std::endl;
                            ss << tabs << "[" << std::endl;
                            for (unsigned int i = 0; i < entitySnapshot.size(); ++i)
                            {
                                const std::string& newlyGenerate = entitySnapshot[i];
                                ss << tabs << "\t(" << i << " => " << newlyGenerate << ")";
                                if (i < entitySnapshot.size() - 1)
                                {
                                    ss << ", ";
                                }
                                ss << std::endl;
                            }
                            ss << tabs << "]" << std::endl;
                        }
                    }
                }
            }
        }
        return ss.str();
    }

    void FileSystemMemoryManager::merge(const FileSystemLookupMemory& info)
    {
        internalMemory.merge(info);
    }
}
