#include "CoreSegmentQueryProcessor.h"

#include <regex>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/ice_conversions.h>
#include <RobotAPI/libraries/armem/core/error.h>


namespace armarx::armem::query_proc
{

    void CoreSegmentQueryProcessor::process(CoreSegment& result, const armem::query::data::CoreSegmentQuery& query, const CoreSegment& coreSegment) const
    {
        if (auto q = dynamic_cast<const armem::query::data::core::All*>(&query))
        {
            process(result, *q, coreSegment);
        }
        else if (auto q = dynamic_cast<const armem::query::data::core::Single*>(&query))
        {
            process(result, *q, coreSegment);
        }
        else if (auto q = dynamic_cast<const armem::query::data::core::Regex*>(&query))
        {
            process(result, *q, coreSegment);
        }
        else
        {
            throw armem::error::UnknownQueryType("core segment", query);
        }
    }


    void CoreSegmentQueryProcessor::process(CoreSegment& result,
                                            const armem::query::data::core::All& query, const CoreSegment& coreSegment) const
    {
        for (const auto& [name, providerSegment] : coreSegment.providerSegments)
        {
            result.addProviderSegment(providerSegmentProcessor.process(query.providerSegmentQueries, providerSegment));
        }
    }

    void CoreSegmentQueryProcessor::process(CoreSegment& result,
                                            const armem::query::data::core::Single& query, const CoreSegment& coreSegment) const
    {
        try
        {
            const ProviderSegment& providerSegment = coreSegment.getProviderSegment(query.providerSegmentName);
            result.addProviderSegment(providerSegmentProcessor.process(query.providerSegmentQueries, providerSegment));
        }
        catch (const error::MissingEntry&)
        {
            // Leave empty.
        }
    }

    void CoreSegmentQueryProcessor::process(CoreSegment& result,
                                            const armem::query::data::core::Regex& query, const CoreSegment& coreSegment) const
    {
        std::regex regex(query.providerSegmentNameRegex);
        for (const auto& [name, providerSegment] : coreSegment.providerSegments)
        {
            if (std::regex_search(providerSegment.name(), regex))
            {
                result.addProviderSegment(providerSegmentProcessor.process(query.providerSegmentQueries, providerSegment));
            }
        }
    }


    data::CoreSegment CoreSegmentQueryProcessor::processToIce(
        const armem::query::data::CoreSegmentQuery& query, const CoreSegment& coreSegment) const
    {
        return toIce<data::CoreSegment>(process(query, coreSegment));
    }

}
