#include "EntityQueryProcessor.h"

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>


namespace armarx::armem::query_proc
{

    void EntityQueryProcessor::process(Entity& result, const armem::query::data::EntityQuery& query, const Entity& entity) const
    {
        if (auto q = dynamic_cast<const armem::query::data::entity::All*>(&query))
        {
            process(result, *q, entity);
        }
        else if (auto q = dynamic_cast<const armem::query::data::entity::Single*>(&query))
        {
            process(result, *q, entity);
        }
        else if (auto q = dynamic_cast<const armem::query::data::entity::TimeRange*>(&query))
        {
            process(result, *q, entity);
        }
        else if (auto q = dynamic_cast<const armem::query::data::entity::IndexRange*>(&query))
        {
            process(result, *q, entity);
        }
        else
        {
            throw armem::error::UnknownQueryType("entity snapshot", query);
        }
    }

    void EntityQueryProcessor::process(Entity& result, const armem::query::data::entity::All& query, const Entity& entity) const
    {
        (void) query;
        // Copy this entitiy and its contents.

        for (const auto& [time, snapshot] : entity)
        {
            addResultSnapshot(result, snapshot);
        }
    }

    void EntityQueryProcessor::process(Entity& result, const armem::query::data::entity::Single& query, const Entity& entity) const
    {
        if (query.timestamp < 0)
        {
            try
            {
                addResultSnapshot(result, entity.getLatestSnapshot());
            }
            catch (const armem::error::EntityHistoryEmpty&)
            {
                // Leave empty.
            }
        }
        else
        {
            Time time = fromIce<Time>(query.timestamp);
            try
            {
                addResultSnapshot(result, entity.getSnapshot(time));
            }
            catch (const armem::error::MissingEntry&)
            {
                // Leave empty.
            }
        }
    }

    void EntityQueryProcessor::process(Entity& result, const armem::query::data::entity::IndexRange& query, const Entity& entity) const
    {
        if (entity.empty())
        {
            return;
        }

        size_t first = negativeIndexSemantics(query.first, entity.history.size());
        size_t last = negativeIndexSemantics(query.last, entity.history.size());

        if (first <= last)
        {
            auto it = entity.begin();
            std::advance(it, first);

            size_t num = last - first + 1;  // +1 to make last inclusive
            for (size_t i = 0; i < num; ++i, ++it)
            {
                addResultSnapshot(result, it);
            }
        }
    }

    void EntityQueryProcessor::process(Entity& result, const armem::query::data::entity::TimeRange& query, const Entity& entity) const
    {
        if (query.minTimestamp <= query.maxTimestamp || query.minTimestamp < 0 || query.maxTimestamp < 0)
        {
            Time min = fromIce<Time>(query.minTimestamp);
            Time max = fromIce<Time>(query.maxTimestamp);
            process(result, min, max, entity, query);
        }
    }

    void EntityQueryProcessor::process(Entity& result, const Time& min, const Time& max, const Entity& entity,
                                       const armem::query::data::EntityQuery& query) const
    {
        (void) query;

        // Returns an iterator pointing to the first element that is not less than (i.e. greater or equal to) key.
        auto begin = min.toMicroSeconds() > 0 ? entity.history.lower_bound(min) : entity.history.begin();
        // Returns an iterator pointing to the first element that is *greater than* key.
        auto end = max.toMicroSeconds() > 0 ? entity.history.upper_bound(max) : entity.history.end();

        for (auto it = begin; it != end && it != entity.history.end(); ++it)
        {
            addResultSnapshot(result, it);
        }
    }

    data::Entity EntityQueryProcessor::processToIce(const armem::query::data::EntityQuery& query, const Entity& entity) const
    {
        return toIce<data::Entity>(process(query, entity));
    }

    size_t EntityQueryProcessor::negativeIndexSemantics(long index, size_t size)
    {
        const size_t max = size > 0 ? size - 1 : 0;
        if (index >= 0)
        {
            return std::clamp<size_t>(static_cast<size_t>(index), 0, max);
        }
        else
        {
            return static_cast<size_t>(std::clamp<long>(static_cast<long>(size) + index, 0, static_cast<long>(max)));
        }
    }

    void EntityQueryProcessor::addResultSnapshot(Entity& result, Entity::ContainerT::const_iterator it) const
    {
        addResultSnapshot(result, it->second);
    }

    void EntityQueryProcessor::addResultSnapshot(Entity& result, const EntitySnapshot& snapshot) const
    {
        bool withData = (dataMode == DataMode::WithData);
        result.addSnapshot(snapshot.copy(withData));
    }
}
