#include "ProviderSegmentQueryProcessor.h"

#include <regex>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>


namespace armarx::armem::query_proc
{

    void ProviderSegmentQueryProcessor::process(ProviderSegment& result, const armem::query::data::ProviderSegmentQuery& query, const ProviderSegment& providerSegment) const
    {
        if (auto q = dynamic_cast<const armem::query::data::provider::All*>(&query))
        {
            process(result, *q, providerSegment);
        }
        else if (auto q = dynamic_cast<const armem::query::data::provider::Single*>(&query))
        {
            process(result, *q, providerSegment);
        }
        else if (auto q = dynamic_cast<const armem::query::data::provider::Regex*>(&query))
        {
            process(result, *q, providerSegment);
        }
        else
        {
            throw armem::error::UnknownQueryType("provider segment", query);
        }
    }


    void ProviderSegmentQueryProcessor::process(ProviderSegment& result,
            const armem::query::data::provider::All& query, const ProviderSegment& providerSegment) const
    {
        for (const auto& [name, entity] : providerSegment.entities)
        {
            result.addEntity(entityProcessor.process(query.entityQueries, entity));
        }
    }

    void ProviderSegmentQueryProcessor::process(ProviderSegment& result,
            const armem::query::data::provider::Single& query, const ProviderSegment& providerSegment) const
    {
        try
        {
            const Entity& entity = providerSegment.getEntity(query.entityName);
            result.addEntity(entityProcessor.process(query.entityQueries, entity));
        }
        catch (const error::MissingEntry&)
        {
            // Leave empty.
        }
    }

    void ProviderSegmentQueryProcessor::process(ProviderSegment& result,
            const armem::query::data::provider::Regex& query, const ProviderSegment& providerSegment) const
    {
        std::regex regex(query.entityNameRegex);
        for (const auto& [name, entity] : providerSegment.entities)
        {
            if (std::regex_search(entity.name(), regex))
            {
                result.addEntity(entityProcessor.process(query.entityQueries, entity));
            }
        }
    }


    data::ProviderSegment ProviderSegmentQueryProcessor::processToIce(
        const armem::query::data::ProviderSegmentQuery& query, const ProviderSegment& providerSegment) const
    {
        return toIce<data::ProviderSegment>(process(query, providerSegment));
    }

}
