#include "TypedDataTreeBuilder.h"

#include <QTreeWidgetItem>

#include <RobotAPI/libraries/armem_gui/TreeWidgetBuilder.h>
#include <RobotAPI/libraries/armem_gui/instance/serialize_path.h>
#include <RobotAPI/libraries/armem_gui/instance/sanitize_typename.h>
#include <RobotAPI/libraries/armem_gui/instance/display_visitors/TypedDataDisplayVisitor.h>


namespace armarx::armem::gui::instance
{

    TypedDataTreeBuilder::TypedDataTreeBuilder()
    {
    }


    void TypedDataTreeBuilder::updateTree(
        QTreeWidgetItem* parent,
        aron::typenavigator::DictNavigator& type,
        aron::datanavigator::DictNavigator& data)
    {
        auto childType = type.getAcceptedType();
        if (childType)
        {
            DictBuilder builder = getDictBuilder();
            builder.setUpdateItemFn([this, &childType, &data](const std::string & key, QTreeWidgetItem * item)
            {
                auto childData = data.getElement(key);
                if (childData)
                {
                    this->update(item, key, *childType, *childData);
                }
                return true;
            });

            builder.updateTree(parent, data.getAllKeys());
        }
    }

    void TypedDataTreeBuilder::updateTree(
        QTreeWidgetItem* parent,
        aron::typenavigator::ObjectNavigator& type,
        aron::datanavigator::DictNavigator& data)
    {
        DictBuilder builder = getDictBuilder();
        builder.setUpdateItemFn([this, &type, &data](const std::string & key, QTreeWidgetItem * item)
        {
            auto childType = type.getMemberType(key);
            auto childData = data.getElement(key);

            if (childType && childData)
            {
                this->update(item, key, *childType, *childData);
            }
            return true;
        });

        builder.updateTree(parent, type.getAllKeys());
    }


    void TypedDataTreeBuilder::updateTree(QTreeWidgetItem* parent,
                                          aron::typenavigator::ListNavigator& type,
                                          aron::datanavigator::ListNavigator& data)
    {
        auto childType = type.getAcceptedType();
        if (childType)
        {
            auto children = data.getChildren();

            ListBuilder builder = getListBuilder();
            builder.setUpdateItemFn([this, &children, &childType](size_t key, QTreeWidgetItem * item)
            {
                auto childData = children.at(key);
                if (childData)
                {
                    this->update(item, std::to_string(key), *childType, *childData);
                }
                return true;
            });

            builder.updateTree(parent, getIndex(children.size()));
        }
    }

    void TypedDataTreeBuilder::updateTree(
        QTreeWidgetItem* parent,
        aron::typenavigator::PairNavigator& type,
        aron::datanavigator::ListNavigator& data)
    {
        ARMARX_CHECK_EQUAL(data.childrenSize(), 2);
        auto childTypes = type.getAcceptedTypes();

        ListBuilder builder = getListBuilder();
        builder.setUpdateItemFn([this, &data, &childTypes](size_t i, QTreeWidgetItem * item)
        {
            auto childType = i == 0 ? childTypes.first : childTypes.second;
            auto childData = data.getElement(static_cast<unsigned int>(i));

            if (childType && childData)
            {
                this->update(item, std::to_string(i), *childType, *childData);
            }
            return true;
        });

        builder.updateTree(parent, getIndex(data.childrenSize()));
    }

    void TypedDataTreeBuilder::updateTree(
        QTreeWidgetItem* parent,
        aron::typenavigator::TupleNavigator& type,
        aron::datanavigator::ListNavigator& data)
    {
        auto childTypes = type.getAcceptedTypes();

        ListBuilder builder = getListBuilder();
        builder.setUpdateItemFn([this, &data, &childTypes](size_t i, QTreeWidgetItem * item)
        {
            auto childType = childTypes.at(i);
            auto childData = data.getElement(static_cast<unsigned int>(i));

            if (childType && childData)
            {
                this->update(item, std::to_string(i), *childType, *childData);
            }
            return true;
        });

        builder.updateTree(parent, getIndex(type.getAcceptedTypes().size()));
    }


    void TypedDataTreeBuilder::update(
        QTreeWidgetItem* item,
        const std::string& key,
        aron::typenavigator::Navigator& type,
        aron::datanavigator::Navigator& data)
    {
        using namespace aron::typenavigator;

        const std::string value = armarx::aron::TypedDataDisplayVisitor::getValue(type, data);
        const std::string typeName = sanitizeTypeName(type.getName());
        setRowTexts(item, key, value, typeName);

        item->setData(columnKey, Qt::UserRole, serializePath(data.getPath()));
        item->setData(columnType, Qt::UserRole, static_cast<int>(type.getDescriptor()));

        if (auto t = dynamic_cast<ObjectNavigator*>(&type))
        {
            _updateTree<aron::datanavigator::DictNavigator>(item, *t, data);
        }
        else if (auto t = dynamic_cast<DictNavigator*>(&type))
        {
            _updateTree<aron::datanavigator::DictNavigator>(item, *t, data);
        }
        else if (auto t = dynamic_cast<ListNavigator*>(&type))
        {
            _updateTree<aron::datanavigator::ListNavigator>(item, *t, data);
        }
        else if (auto t = dynamic_cast<PairNavigator*>(&type))
        {
            _updateTree<aron::datanavigator::ListNavigator>(item, *t, data);
        }
        else if (auto t = dynamic_cast<TupleNavigator*>(&type))
        {
            _updateTree<aron::datanavigator::ListNavigator>(item, *t, data);
        }
    }

    template <class DataT, class TypeT>
    void TypedDataTreeBuilder::_updateTree(QTreeWidgetItem* item, TypeT& type, aron::datanavigator::Navigator& data)
    {
        DataT& dataCast = dynamic_cast<DataT&>(data);
        updateTree(item, type, dataCast);
    }

}
