#include "Segment.h"

#include <RobotAPI/libraries/ArmarXObjects/aron_conversions.h>
#include <RobotAPI/libraries/armem_objects/aron_conversions.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/time/TimeUtil.h>

#include <filesystem>


namespace armarx::armem::server::obj::clazz
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter) :
        iceMemory(memoryToIceAdapter)
    {
        Logging::setTag("ClassSegment");
    }

    void Segment::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(p.coreSegmentName, prefix + "CoreSegmentName", "Name of the object clazz core segment.");
        defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of object poses history (-1 for infinite).");

        defs->optional(p.objectsPackage, prefix + "ObjectsPackgage", "Name of the objects package to load from.");
        defs->optional(p.loadFromObjectsPackage, prefix + "LoadFromObjectsPackage",
                       "If true, load the objects from the objects package on startup.");
    }

    void Segment::init()
    {
        ARMARX_CHECK_NOT_NULL(iceMemory.memory);

        coreSegment = &iceMemory.memory->addCoreSegment(p.coreSegmentName, arondto::ObjectClass::toInitialAronType());
        coreSegment->setMaxHistorySize(p.maxHistorySize);

        if (p.loadFromObjectsPackage)
        {
            loadByObjectFinder(p.objectsPackage);
        }
    }


    armem::CoreSegment& Segment::getCoreSegment()
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        return *coreSegment;
    }

    const armem::CoreSegment& Segment::getCoreSegment() const
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        return *coreSegment;
    }

    void Segment::loadByObjectFinder(const std::string& objectsPackage)
    {
        loadByObjectFinder(ObjectFinder(objectsPackage));
    }

    void Segment::loadByObjectFinder(const ObjectFinder& finder)
    {
        this->objectFinder = finder;
        loadByObjectFinder();
    }

    void Segment::loadByObjectFinder()
    {
        const Time now = TimeUtil::GetTime();

        const bool checkPaths = false;
        std::vector<ObjectInfo> infos = objectFinder.findAllObjects(checkPaths);

        const MemoryID providerID = coreSegment->id().withProviderSegmentName(objectFinder.getPackageName());
        coreSegment->addProviderSegment(providerID.providerSegmentName);

        ARMARX_INFO << "Loading up to " << infos.size() << " object classes from '"
                    << objectFinder.getPackageName() << "' ...";
        Commit commit;
        for (ObjectInfo& info : infos)
        {
            info.setLogError(false);

            EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = providerID.withEntityName(info.id().str());
            update.timeArrived = update.timeCreated = update.timeSent = now;

            arondto::ObjectClass objectClass = objectClassFromInfo(info);
            update.instancesData =
            {
                objectClass.toAron()
            };
        }
        ARMARX_INFO << "Loaded " << commit.updates.size() << " object classes from '"
                    << objectFinder.getPackageName() << "'.";
        iceMemory.commit(commit);
    }

    arondto::ObjectClass Segment::objectClassFromInfo(const ObjectInfo& info) const
    {
        namespace fs = std::filesystem;

        arondto::ObjectClass data;

        toAron(data.id, info.id());

        if (fs::is_regular_file(info.simoxXML().absolutePath))
        {
            toAron(data.simoxXmlPath, info.simoxXML());
        }
        if (fs::is_regular_file(info.wavefrontObj().absolutePath))
        {
            toAron(data.meshObjPath, info.wavefrontObj());
        }
        if (fs::is_regular_file(info.file(".wrl").absolutePath))
        {
            toAron(data.meshWrlPath, info.file(".wrl"));
        }

        if (auto aabb = info.loadAABB())
        {
            toAron(data.aabb, aabb.value());
        }
        if (auto oobb = info.loadOOBB())
        {
            toAron(data.oobb, oobb.value());
        }

        if (auto recogNames = info.loadRecognizedNames())
        {
            data.names.recognizedNames = recogNames.value();
        }
        if (auto spokenNames = info.loadSpokenNames())
        {
            data.names.spokenNames = spokenNames.value();
        }

        return data;
    }


    void Segment::RemoteGui::setup(const Segment& data)
    {
        using namespace armarx::RemoteGui::Client;

        maxHistorySize.setValue(std::max(1, int(data.p.maxHistorySize)));
        maxHistorySize.setRange(1, 1e6);
        infiniteHistory.setValue(data.p.maxHistorySize == -1);

        GridLayout grid;
        int row = 0;
        grid.add(Label("Max History Size"), {row, 0}).add(maxHistorySize, {row, 1});
        row++;
        grid.add(Label("Infinite History Size"), {row, 0}).add(infiniteHistory, {row, 1});
        row++;

        group.setLabel("Data");
        group.addChild(grid);
    }

    void Segment::RemoteGui::update(Segment& data)
    {
        if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged())
        {
            data.p.maxHistorySize = infiniteHistory.getValue() ? -1 : maxHistorySize.getValue();
            if (data.coreSegment)
            {
                data.coreSegment->setMaxHistorySize(long(data.p.maxHistorySize));
            }
        }
    }

}
