#include "Segment.h"

#include <RobotAPI/libraries/armem_objects/aron_conversions.h>

#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectFinder.h>
#include <RobotAPI/libraries/ArmarXObjects/aron_conversions.h>
#include <RobotAPI/libraries/ArmarXObjects/ice_conversions.h>
#include <RobotAPI/libraries/ArmarXObjects/aron/ObjectPose.aron.generated.h>
#include <RobotAPI/libraries/armem/client/Writer.h>
#include <RobotAPI/libraries/armem/core/Visitor.h>
#include <RobotAPI/libraries/aron/converter/eigen/EigenConverter.h>


#include <ArmarXCore/core/time/TimeUtil.h>

#include <sstream>


namespace armarx::armem::server::obj::instance
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter) :
        iceMemory(memoryToIceAdapter)
    {
        Logging::setTag("InstanceSegment");

        oobbCache.setFetchFn([this](const ObjectID & id) -> std::optional<simox::OrientedBoxf>
        {
            // Try to get OOBB from repository.
            if (std::optional<ObjectInfo> objectInfo = objectFinder.findObject(id))
            {
                try
                {
                    return objectInfo->loadOOBB();
                }
                catch (const std::ios_base::failure& e)
                {
                    // Give up - no OOBB information.
                    ARMARX_WARNING << "Could not get OOBB of object " << id << ".\n- " << e.what();
                    return std::nullopt;
                }
            }
            else
            {
                return std::nullopt;
            }
        });

        classNameToDatasetCache.setFetchFn([this](const std::string & className)
        {
            std::optional<ObjectInfo> objectInfo = objectFinder.findObject(className);
            return objectInfo ? objectInfo->dataset() : "";
        });
    }

    void Segment::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(p.coreSegmentName, prefix + "CoreSegmentName", "Name of the object instance core segment.");
        defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of object poses history (-1 for infinite).");
        defs->optional(p.discardSnapshotsWhileAttached, prefix + "DiscardSnapshotsWhileAttached",
                       "If true, no new snapshots are stored while an object is attached to a robot node.\n"
                       "If false, new snapshots are stored, but the attachment is kept in the new snapshots.");

        decay.defineProperties(defs, prefix + "decay.");
    }

    void Segment::init()
    {
        ARMARX_CHECK_NOT_NULL(iceMemory.memory);

        coreSegment = &iceMemory.memory->addCoreSegment(p.coreSegmentName, arondto::ObjectInstance::toInitialAronType());
        coreSegment->setMaxHistorySize(p.maxHistorySize);
    }


    Segment::CommitStats Segment::commitObjectPoses(
        const std::string& providerName,
        const objpose::data::ProvidedObjectPoseSeq& providedPoses,
        std::optional<armem::Time> discardUpdatesUntil)
    {
        CommitStats stats;

        // Build new poses.
        objpose::ObjectPoseSeq newObjectPoses;
        stats.numUpdated = 0;
        for (const objpose::data::ProvidedObjectPose& provided : providedPoses)
        {
            const IceUtil::Time timestamp = IceUtil::Time::microSeconds(provided.timestampMicroSeconds);

            // Check whether we have an old snapshot for this object.
            std::optional<objpose::ObjectPose> previousPose;
            const armem::Entity* entity = findObjectEntity(armarx::fromIce(provided.objectID), providerName);
            if (entity)
            {
                const arondto::ObjectInstance data = getLatestInstanceData(*entity);

                previousPose = objpose::ObjectPose();
                fromAron(data, *previousPose);
            }

            bool discard = false;
            if (discardUpdatesUntil && timestamp < discardUpdatesUntil.value())
            {
                // Dicard updates temporarily (e.g. due to head movement).
                discard = true;
            }
            else if (previousPose)
            {
                if (p.discardSnapshotsWhileAttached && previousPose->attachment)
                {
                    // Discard update due to active attachemnt.
                    discard = true;
                }
                else if (timestamp == previousPose->timestamp)
                {
                    // Discard update as it is not new.
                    discard = true;
                }
            }

            if (!discard)
            {
                // Update the entity.
                stats.numUpdated++;

                objpose::ObjectPose& newPose = newObjectPoses.emplace_back();
                newPose.fromProvidedPose(provided, robot);

                if (previousPose && previousPose->attachment)
                {
                    // Keep current attachment.
                    ARMARX_CHECK(!p.discardSnapshotsWhileAttached);
                    newPose.attachment = previousPose->attachment;
                }

                if (newPose.objectID.dataset().empty())
                {
                    // Try to find the data set.
                    const std::string dataset = classNameToDatasetCache.get(newPose.objectID.className());
                    if (!dataset.empty())
                    {
                        newPose.objectID = { dataset, newPose.objectID.className(), newPose.objectID.instanceName() };
                    }
                }
                if (!provided.localOOBB)
                {
                    // Try to load oobb from disk.
                    newPose.localOOBB = getObjectOOBB(newPose.objectID);
                }
            }
        }

        commitObjectPoses(providerName, newObjectPoses);

        return stats;
    }

    void Segment::commitObjectPoses(const std::string& providerName, const ObjectPoseSeq& objectPoses)
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);

        // Update memory.
        const armem::MemoryID providerSegmentID = coreSegment->id().withProviderSegmentName(providerName);
        if (!coreSegment->hasProviderSegment(providerSegmentID.providerSegmentName))
        {
            coreSegment->addProviderSegment(providerSegmentID.providerSegmentName);
        }

        armem::Commit commit;
        for (const objpose::ObjectPose& pose : objectPoses)
        {
            armem::EntityUpdate& update = commit.updates.emplace_back();
            {
                update.entityID = providerSegmentID.withEntityName(pose.objectID.str());
                update.timeArrived = TimeUtil::GetTime();
                update.timeCreated = pose.timestamp;
                update.confidence = pose.confidence;

                arondto::ObjectInstance dto;
                toAron(dto, pose);
                update.instancesData.push_back(dto.toAron());
            }

            iceMemory.commit(commit);
        }
    }


    armem::CoreSegment& Segment::getCoreSegment()
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        return *coreSegment;
    }

    const armem::CoreSegment& Segment::getCoreSegment() const
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        return *coreSegment;
    }

    objpose::ObjectPoseSeq Segment::getObjectPoses(IceUtil::Time now)
    {
        ObjectPoseSeq objectPoses = getLatestObjectPoses();
        updateObjectPoses(objectPoses, now);
        return filterObjectPoses(objectPoses);
    }


    objpose::ObjectPoseSeq Segment::getObjectPosesByProvider(
        const std::string& providerName,
        IceUtil::Time now)
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        ObjectPoseSeq objectPoses = getLatestObjectPoses(coreSegment->getProviderSegment(providerName));
        updateObjectPoses(objectPoses, now);
        return filterObjectPoses(objectPoses);
    }

    armem::Entity* Segment::findObjectEntity(const ObjectID& objectID, const std::string& providerName)
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        armem::MemoryID entityID = armem::MemoryID().withEntityName(objectID.str());
        if (providerName.empty())
        {
            for (auto& [_, prov] : *coreSegment)
            {
                if (prov.hasEntity(entityID.entityName))
                {
                    return &prov.getEntity(entityID);
                }
            }
            return nullptr;
        }
        else
        {
            entityID.providerSegmentName = providerName;
            if (coreSegment->hasProviderSegment(providerName))
            {
                armem::ProviderSegment& prov = coreSegment->getProviderSegment(providerName);
                return prov.hasEntity(entityID.entityName) ? &prov.getEntity(entityID) : nullptr;
            }
            else
            {
                return nullptr;
            }
        }
    }


    void Segment::updateObjectPoses(ObjectPoseSeq& objectPoses, IceUtil::Time now)
    {
        bool agentSynchronized = false;

        for (ObjectPose& objectPose : objectPoses)
        {
            updateObjectPose(objectPose, now, robot, agentSynchronized);
        }
    }


    void Segment::updateObjectPoses(
        ObjectPoseSeq& objectPoses,
        IceUtil::Time now,
        VirtualRobot::RobotPtr agent,
        bool& agentSynchronized) const
    {
        for (ObjectPose& pose : objectPoses)
        {
            updateObjectPose(pose, now, agent, agentSynchronized);
        }
    }


    void Segment::updateObjectPose(
        ObjectPose& objectPose,
        IceUtil::Time now,
        VirtualRobot::RobotPtr agent,
        bool& agentSynchronized) const
    {
        updateAttachement(objectPose, agent, agentSynchronized);

        if (decay.enabled)
        {
            decay.updateConfidence(objectPose, now);
        }
    }


    objpose::ObjectPoseSeq Segment::filterObjectPoses(const ObjectPoseSeq& objectPoses) const
    {
        ObjectPoseSeq result;
        for (const ObjectPose& objectPose : objectPoses)
        {
            if (!(decay.enabled && objectPose.confidence < decay.removeObjectsBelowConfidence))
            {
                result.push_back(objectPose);
            }
        }
        return result;
    }


    void Segment::updateAttachement(
        ObjectPose& objectPose, VirtualRobot::RobotPtr agent, bool& synchronized) const
    {
        if (!objectPose.attachment)
        {
            // No attachment, nothing to do.
            return;
        }
        ARMARX_CHECK(objectPose.attachment);

        if (!synchronized)  // Synchronize only once.
        {
            RemoteRobot::synchronizeLocalClone(agent, robotStateComponent);
            synchronized = true;
        }
        objectPose.updateAttached(agent);
    }

    objpose::ObjectPoseSeq Segment::getLatestObjectPoses() const
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);
        return getLatestObjectPoses(*coreSegment);
    }

    objpose::ObjectPoseSeq Segment::getLatestObjectPoses(const armem::CoreSegment& coreSeg)
    {
        ObjectPoseSeq result;
        getLatestObjectPoses(coreSeg, result);
        return result;
    }

    objpose::ObjectPoseSeq Segment::getLatestObjectPoses(const armem::ProviderSegment& provSeg)
    {
        ObjectPoseSeq result;
        getLatestObjectPoses(provSeg, result);
        return result;
    }

    objpose::ObjectPose Segment::getLatestObjectPose(const armem::Entity& entity)
    {
        ObjectPose result;
        getLatestObjectPose(entity, result);
        return result;
    }

    void Segment::getLatestObjectPoses(const armem::CoreSegment& coreSeg, ObjectPoseSeq& out)
    {
        for (const auto& [_, provSegment] : coreSeg)
        {
            getLatestObjectPoses(provSegment, out);
        }
    }

    void Segment::getLatestObjectPoses(const armem::ProviderSegment& provSegment, ObjectPoseSeq& out)
    {
        for (const auto& [_, entity] : provSegment)
        {
            if (!entity.empty())
            {
                ObjectPose& pose = out.emplace_back();
                getLatestObjectPose(entity, pose);
            }
        }
    }

    void Segment::getLatestObjectPose(const armem::Entity& entity, ObjectPose& out)
    {
        for (const armem::EntityInstance& instance : entity.getLatestSnapshot())
        {
            arondto::ObjectInstance dto;
            dto.fromAron(instance.data());

            fromAron(dto, out);
        }
    }


    arondto::ObjectInstance Segment::getLatestInstanceData(const armem::Entity& entity)
    {
        ARMARX_CHECK_GREATER_EQUAL(entity.size(), 1);
        const armem::EntitySnapshot& snapshot = entity.getLatestSnapshot();

        ARMARX_CHECK_EQUAL(snapshot.size(), 1);
        const armem::EntityInstance& instance = snapshot.getInstance(0);

        arondto::ObjectInstance data;
        data.fromAron(instance.data());

        return data;
    }

    std::optional<simox::OrientedBoxf> Segment::getObjectOOBB(const ObjectID& id)
    {
        return oobbCache.get(id);
    }

    objpose::ProviderInfo Segment::getProviderInfo(const std::string& providerName)
    {
        try
        {
            return providers.at(providerName);
        }
        catch (const std::out_of_range&)
        {
            std::stringstream ss;
            ss << "No provider with name '" << providerName << "' available.\n";
            ss << "Available are:\n";
            for (const auto& [name, _] : providers)
            {
                ss << "- '" << name << "'\n";
            }
            throw std::out_of_range(ss.str());
        }
    }



    objpose::AttachObjectToRobotNodeOutput
    Segment::attachObjectToRobotNode(const objpose::AttachObjectToRobotNodeInput& input)
    {
        const armem::Time now = armem::Time::now();

        objpose::AttachObjectToRobotNodeOutput output;
        output.success = false;  // We are not successful until proven otherwise.

        ObjectID objectID = armarx::fromIce(input.objectID);

        if (input.agentName != "" && input.agentName != this->robot->getName())
        {
            ARMARX_WARNING << "Tried to attach object " << objectID << " to unknown agent '" << input.agentName << "'."
                           << "\n(You can leave the agent name empty if there is only one agent.)\n"
                           << "\nKnown agents: " << std::vector<std::string> {this->robot->getName()};
            return output;
        }
        VirtualRobot::RobotPtr agent = this->robot;

        if (!agent->hasRobotNode(input.frameName))
        {
            ARMARX_WARNING << "Tried to attach object " << objectID << " to unknown node '" << input.frameName
                           << "' of agent '" << agent->getName() << "'.";
            return output;
        }
        std::string frameName = input.frameName;


        // Find object pose (provider name can be empty).
        armem::Entity* objectEntity = this->findObjectEntity(objectID, input.providerName);
        if (!objectEntity || objectEntity->empty())
        {
            ARMARX_WARNING << "Tried to attach object " << objectID << " to node '" << frameName
                           << "' of agent '" << agent->getName() << "', but object is currently not provided.";
            return output;
        }
        arondto::ObjectInstance data = getLatestInstanceData(*objectEntity);

        objpose::ObjectAttachmentInfo info;
        info.agentName = agent->getName();
        info.frameName = frameName;

        if (input.poseInFrame)
        {
            info.poseInFrame = PosePtr::dynamicCast(input.poseInFrame)->toEigen();
        }
        else
        {
            RemoteRobot::synchronizeLocalClone(agent, robotStateComponent);

            armarx::FramedPose framed(data.pose.objectPoseGlobal, armarx::GlobalFrame, agent->getName());
            if (frameName == armarx::GlobalFrame)
            {
                info.poseInFrame = framed.toGlobalEigen(this->robot);
            }
            else
            {
                framed.changeFrame(this->robot, info.frameName);
                info.poseInFrame = framed.toEigen();
            }
        }

        // Store attachment in new entity snapshot.
        {
            armem::EntityUpdate update;
            update.entityID = objectEntity->id();
            update.timeCreated = now;
            {
                arondto::ObjectInstance updated = data;
                toAron(updated.pose.attachment, info);
                updated.pose.attachmentValid = true;
                update.instancesData = { updated.toAron() };
            }
            objectEntity->update(update);
        }

        ARMARX_INFO << "Attached object " << objectID << " by provider '" << data.pose.providerName << "' "
                    << "to node '" << info.frameName << "' of agent '" << info.agentName << "'.\n"
                    << "Object pose in frame: \n" << info.poseInFrame;

        output.success = true;
        output.attachment = new objpose::data::ObjectAttachmentInfo();
        output.attachment->frameName = info.frameName;
        output.attachment->agentName = info.agentName;
        output.attachment->poseInFrame = new Pose(info.poseInFrame);

        return output;
    }

    objpose::DetachObjectFromRobotNodeOutput Segment::detachObjectFromRobotNode(
        const objpose::DetachObjectFromRobotNodeInput& input)
    {
        const armem::Time now = armem::Time::now();

        ObjectID objectID = armarx::fromIce(input.objectID);
        std::string providerName = input.providerName;

        std::optional<objpose::arondto::ObjectAttachmentInfo> attachment;
        {
            // Remove from latest pose (if it was cached).
            // Find object pose (provider name can be empty).
            armem::Entity* entity = this->findObjectEntity(objectID, input.providerName);
            if (entity)
            {
                const arondto::ObjectInstance data = getLatestInstanceData(*entity);
                if (data.pose.attachmentValid)
                {
                    attachment = data.pose.attachment;

                    // Store non-attached pose in new snapshot.
                    storeDetachedSnapshot(*entity, data, now, input.commitAttachedPose);
                }

                if (providerName.empty())
                {
                    providerName = data.pose.providerName;
                }
            }
        }

        objpose::DetachObjectFromRobotNodeOutput output;
        output.wasAttached = bool(attachment);
        if (attachment)
        {
            ARMARX_INFO << "Detached object " << objectID << " by provider '" << providerName << "' from robot node '"
                        << attachment->frameName << "' of agent '" << attachment->agentName << "'.";
        }
        else
        {
            ARMARX_INFO << "Tried to detach object " << objectID << " by provider '" << providerName << "' "
                        << "from robot node, but it was not attached.";
        }

        return output;
    }


    struct DetachVisitor : public armem::Visitor
    {
        Segment& owner;
        armem::Time now;
        bool commitAttachedPose;

        int numDetached = 0;

        DetachVisitor(Segment& owner, armem::Time now, bool commitAttachedPose) :
            owner(owner), now(now), commitAttachedPose(commitAttachedPose)
        {
        }

        virtual bool visitEnter(armem::Entity& entity) override;
    };

    bool DetachVisitor::visitEnter(armem::Entity& entity)
    {
        const arondto::ObjectInstance data = owner.getLatestInstanceData(entity);
        if (data.pose.attachmentValid)
        {
            numDetached++;
            // Store non-attached pose in new snapshot.
            owner.storeDetachedSnapshot(entity, data, now, commitAttachedPose);
        }

        return false; // Stop descending.
    }


    objpose::DetachAllObjectsFromRobotNodesOutput Segment::detachAllObjectsFromRobotNodes(
        const objpose::DetachAllObjectsFromRobotNodesInput& input)
    {
        ARMARX_CHECK_NOT_NULL(coreSegment);

        const armem::Time now = armem::Time::now();

        DetachVisitor visitor(*this, now, input.commitAttachedPose);
        visitor.applyTo(*coreSegment);

        objpose::DetachAllObjectsFromRobotNodesOutput output;
        output.numDetached = visitor.numDetached;

        ARMARX_INFO << "Detached all objects (" << output.numDetached << ") from robot nodes.";

        return output;
    }

    void Segment::storeDetachedSnapshot(
        armem::Entity& entity,
        const arondto::ObjectInstance& data,
        armem::Time now,
        bool commitAttachedPose)
    {
        armem::EntityUpdate update;
        update.entityID = entity.id();
        update.timeCreated = now;
        {
            arondto::ObjectInstance updated;
            if (commitAttachedPose && data.pose.attachmentValid)
            {
                ObjectPose objectPose;
                fromAron(data, objectPose);

                bool agentSynchronized = false;
                updateAttachement(objectPose, robot, agentSynchronized);

                objectPose.attachment = std::nullopt;
                toAron(updated, objectPose);
            }
            else
            {
                updated = data;
                updated.pose.attachmentValid = false;
                toAron(updated.pose.attachment, objpose::ObjectAttachmentInfo{});
            }

            update.instancesData = { updated.toAron() };
        }
        entity.update(update);
    }



    void Segment::RemoteGui::setup(const Segment& data)
    {
        using namespace armarx::RemoteGui::Client;

        maxHistorySize.setValue(std::max(1, int(data.p.maxHistorySize)));
        maxHistorySize.setRange(1, 1e6);
        infiniteHistory.setValue(data.p.maxHistorySize == -1);
        discardSnapshotsWhileAttached.setValue(data.p.discardSnapshotsWhileAttached);

        GridLayout grid;
        int row = 0;
        grid.add(Label("Max History Size"), {row, 0}).add(maxHistorySize, {row, 1});
        row++;
        grid.add(Label("Infinite History Size"), {row, 0}).add(infiniteHistory, {row, 1});
        row++;
        grid.add(Label("Discard Snapshots while Attached"), {row, 0}).add(discardSnapshotsWhileAttached, {row, 1});
        row++;

        group.setLabel("Data");
        group.addChild(grid);
    }

    void Segment::RemoteGui::update(Segment& data)
    {
        if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged())
        {
            data.p.maxHistorySize = infiniteHistory.getValue() ? -1 : maxHistorySize.getValue();
            if (data.coreSegment)
            {
                data.coreSegment->setMaxHistorySize(long(data.p.maxHistorySize));
            }
        }

        data.p.discardSnapshotsWhileAttached = discardSnapshotsWhileAttached.getValue();
    }

}
