#pragma once

#include <map>
#include <string>
#include <optional>

#include <SimoxUtility/caching/CacheMap.h>
#include <SimoxUtility/shapes/OrientedBox.h>

#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/interface/core/RobotState.h>
#include <RobotAPI/interface/objectpose/ObjectPoseStorageInterface.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectID.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectPose.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectFinder.h>

#include <RobotAPI/libraries/armem/core/Memory.h>
#include <RobotAPI/libraries/armem/server/MemoryToIceAdapter.h>
#include <RobotAPI/libraries/armem_objects/aron/ObjectInstance.aron.generated.h>

#include "Decay.h"


namespace armarx::armem::server::obj::instance
{

    class Segment : public armarx::Logging
    {
    public:

        struct CommitStats
        {
            int numUpdated = 0;
        };
        using ObjectPose = objpose::ObjectPose;
        using ObjectPoseSeq = objpose::ObjectPoseSeq;


    public:

        Segment(armem::server::MemoryToIceAdapter& iceMemory);


        void defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix = "");

        void init();


        CommitStats commitObjectPoses(
            const std::string& providerName,
            const objpose::data::ProvidedObjectPoseSeq& providedPoses,
            std::optional<armem::Time> discardUpdatesUntil = std::nullopt);
        void commitObjectPoses(const std::string& providerName, const ObjectPoseSeq& objectPoses);


        armem::CoreSegment& getCoreSegment();
        const armem::CoreSegment& getCoreSegment() const;


        objpose::ObjectPoseSeq getObjectPoses(IceUtil::Time now);
        objpose::ObjectPoseSeq getObjectPosesByProvider(const std::string& providerName, IceUtil::Time now);

        armem::Entity* findObjectEntity(const ObjectID& objectID, const std::string& providerName = "");
        std::optional<simox::OrientedBoxf> getObjectOOBB(const ObjectID& id);

        objpose::ProviderInfo getProviderInfo(const std::string& providerName);

        objpose::AttachObjectToRobotNodeOutput attachObjectToRobotNode(const objpose::AttachObjectToRobotNodeInput& input);
        objpose::DetachObjectFromRobotNodeOutput detachObjectFromRobotNode(const objpose::DetachObjectFromRobotNodeInput& input);
        objpose::DetachAllObjectsFromRobotNodesOutput detachAllObjectsFromRobotNodes(const objpose::DetachAllObjectsFromRobotNodesInput& input);


        /**
         * @brief If the object is attached to a robot node, update it according to the current robot state.
         *
         * If there is no attachement info in `objectPose` itself, the internal data
         * structure `attachments` is queried. If an attachment is found there,
         * it is written into the given `objectPose` (thus, it is "cached" in the
         * info `objectPose`).
         *
         * @param synchronized Indicates whether the agent is already synchronized to the current time.
         */
        void updateAttachement(ObjectPose& objectPose, VirtualRobot::RobotPtr agent,
                               bool& synchronized) const;


        static ObjectPoseSeq getLatestObjectPoses(const armem::CoreSegment& coreSeg);
        static ObjectPoseSeq getLatestObjectPoses(const armem::ProviderSegment& provSeg);
        static ObjectPose getLatestObjectPose(const armem::Entity& entity);

        static void getLatestObjectPoses(const armem::CoreSegment& coreSeg, ObjectPoseSeq& out);
        static void getLatestObjectPoses(const armem::ProviderSegment& provSeg, ObjectPoseSeq& out);
        static void getLatestObjectPose(const armem::Entity& entity, ObjectPose& out);

        static arondto::ObjectInstance getLatestInstanceData(const armem::Entity& entity);


    private:

        ObjectPoseSeq getLatestObjectPoses() const;

        void updateObjectPoses(
            ObjectPoseSeq& objectPoses,
            IceUtil::Time now);
        void updateObjectPoses(
            ObjectPoseSeq& objectPoses,
            IceUtil::Time now,
            VirtualRobot::RobotPtr agent,
            bool& agentSynchronized
        ) const;
        void updateObjectPose(
            ObjectPose& objectPose,
            IceUtil::Time now,
            VirtualRobot::RobotPtr agent,
            bool& agentSynchronized
        ) const;


        ObjectPoseSeq filterObjectPoses(const ObjectPoseSeq& objectPoses) const;


        void storeDetachedSnapshot(
            armem::Entity& entity,
            const arondto::ObjectInstance& data,
            armem::Time now,
            bool commitAttachedPose);


        friend struct DetachVisitor;


    public:

        RobotStateComponentInterfacePrx robotStateComponent;
        VirtualRobot::RobotPtr robot;

        objpose::ProviderInfoMap providers;


        ObjectFinder objectFinder;

        /// Decay model.
        Decay decay;


    private:

        armem::server::MemoryToIceAdapter& iceMemory;

        armem::CoreSegment* coreSegment = nullptr;


        struct Properties
        {
            std::string coreSegmentName = "Instance";
            long maxHistorySize = -1;
            bool discardSnapshotsWhileAttached = true;
        };
        Properties p;


        /// Caches results of attempts to retrieve the OOBB from ArmarXObjects.
        simox::caching::CacheMap<ObjectID, std::optional<simox::OrientedBoxf>> oobbCache;

        /// Class name -> dataset name.
        simox::caching::CacheMap<std::string, std::string> classNameToDatasetCache;


    public:

        struct RemoteGui
        {
            armarx::RemoteGui::Client::GroupBox group;

            armarx::RemoteGui::Client::IntSpinBox maxHistorySize;
            armarx::RemoteGui::Client::CheckBox infiniteHistory;
            armarx::RemoteGui::Client::CheckBox discardSnapshotsWhileAttached;

            void setup(const Segment& data);
            void update(Segment& data);
        };

    };

}
