/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "Converter.h"

#include <RobotAPI/libraries/aron/core/Debug.h>



namespace armarx::aron::dataIO
{
    void Converter::ConvertFromReader(ReaderInterface& reader, WriterInterface& writer, const aron::typenavigator::NavigatorPtr& expectedStructure)
    {
        type::Descriptor t_desc = expectedStructure != nullptr ? expectedStructure->getDescriptor() : type::eUnknown;
        data::Descriptor desc = reader.getTypeOfNext(t_desc);
        switch (desc)
        {
            case data::eDict:
            {
                int elements = reader.readStartDict();
                writer.writeStartDict();

                typenavigator::NavigatorPtr childType = nullptr;
                if (t_desc == type::eDict)
                {
                    auto t = typenavigator::DictNavigator::DynamicCast(expectedStructure);
                    childType = t->getAcceptedType();
                }

                for (int i = 0; i < elements; ++i)
                {
                    std::string key = reader.readKey();
                    writer.writeKey(key);

                    if (t_desc == type::eObject)
                    {
                        auto t = typenavigator::ObjectNavigator::DynamicCast(expectedStructure);
                        childType = t->getMemberType(key);
                    }

                    Converter::ConvertFromReader(reader, writer, childType);
                }

                writer.writeEndDict();
                reader.readEndDict();

                break;
            }
            case data::eList:
            {
                int elements = reader.readStartList();
                writer.writeStartList();

                typenavigator::NavigatorPtr childType = nullptr;
                if (t_desc == type::eList)
                {
                    auto t = typenavigator::ListNavigator::DynamicCast(expectedStructure);
                    childType = t->getAcceptedType();
                }

                for (int i = 0; i < elements; ++i)
                {
                    if (t_desc == type::eObject)
                    {
                        auto t = typenavigator::TupleNavigator::DynamicCast(expectedStructure);
                        childType = t->getAcceptedType(i);
                    }

                    Converter::ConvertFromReader(reader, writer, childType);
                }

                writer.writeEndList();
                reader.readEndList();
                break;
            }
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case data::e##upperType: \
{ \
    auto [dims, type] = reader.readStart##upperType(); \
    int elements = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()); \
    std::vector<unsigned char> data(elements); \
    reader.readEnd##upperType(data.data()); \
    \
    writer.write##upperType(dims, type, data.data()); \
    break; \
}

            HANDLE_COMPLEX_DATA
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case data::e##upperType: \
{ \
    lowerType val = reader.readPrimitive<lowerType>(); \
    writer.writePrimitive(val); \
    break; \
}

            HANDLE_PRIMITIVE_DATA
#undef RUN_ARON_MACRO

            default:
            {
                throw error::DescriptorNotValidException("LegacyAronDataReader", "SetupReaderAndGetResult", "Data-Type could not be resolved", desc);
            }
        }
    }
}
