/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "Visitor.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Resolver.h>



namespace armarx::aron::dataIO
{

    void Visitor::SetupWriterFromAronDataPtr(WriterInterface& writer, const datanavigator::NavigatorPtr& aron)
    {
        SetupWriterFromAronDataPtr(writer, aron->getResult());
    }

    void Visitor::SetupWriterFromAronDataPtr(WriterInterface& writer, const data::AronDataPtr& aron)
    {
        data::Descriptor desc = Resolver::GetDescriptor(aron);
        switch (desc)
        {
            case data::eDict:
            {
                data::AronDictPtr casted = data::AronDictPtr::dynamicCast(aron);
                writer.writeStartDict();
                for (const auto& [key, value] : casted->elements)
                {
                    writer.writeKey(key);
                    Visitor::SetupWriterFromAronDataPtr(writer, value);
                }
                writer.writeEndDict();
                break;
            }
            case data::eList:
            {
                data::AronListPtr casted = data::AronListPtr::dynamicCast(aron);
                writer.writeStartList();
                for (const auto& value : casted->elements)
                {
                    Visitor::SetupWriterFromAronDataPtr(writer, value);
                }
                writer.writeEndList();
                break;
            }
            case data::eNDArray:
            {
                data::AronNDArrayPtr casted = data::AronNDArrayPtr::dynamicCast(aron);
                writer.writeNDArray(casted->dimensions, casted->type, casted->data.data());
                break;
            }
            case data::eInt:
            {
                data::AronIntPtr casted = data::AronIntPtr::dynamicCast(aron);
                writer.writePrimitive(casted->value);
                break;
            }
            case data::eLong:
            {
                data::AronLongPtr casted = data::AronLongPtr::dynamicCast(aron);
                writer.writePrimitive(casted->value);
                break;
            }
            case data::eFloat:
            {
                data::AronFloatPtr casted = data::AronFloatPtr::dynamicCast(aron);
                writer.writePrimitive(casted->value);
                break;
            }
            case data::eDouble:
            {
                data::AronDoublePtr casted = data::AronDoublePtr::dynamicCast(aron);
                writer.writePrimitive(casted->value);
                break;
            }
            case data::eString:
            {
                data::AronStringPtr casted = data::AronStringPtr::dynamicCast(aron);
                writer.writePrimitive(casted->value);
                break;
            }
            case data::eBool:
            {
                data::AronBoolPtr casted = data::AronBoolPtr::dynamicCast(aron);
                writer.writePrimitive(casted->value);
                break;
            }
            default:
            {
                throw error::DescriptorNotValidException("LegacyAronDataWriter", "SetupWriterFromAronDataPtr", "Data-Type could not be resolved. The ice_id of the input was: " + aron->ice_id(), desc);
            }
        }
    }
}
