/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "Converter.h"

namespace armarx::aron::typeIO
{
    void Converter::ConvertFromReader(ReaderInterface& reader, WriterInterface& writer)
    {
        type::Descriptor desc = reader.getTypeOfNext();
        switch (desc)
        {
            case type::eObject:
            {
                const auto [name, elements] = reader.readStartObject();
                writer.writeStartObject(name);

                for (int i = 0; i < elements; ++i)
                {
                    std::string key = reader.readKey();
                    writer.writeKey(key);
                    Converter::ConvertFromReader(reader, writer);
                }

                writer.writeEndObject();
                reader.readEndObject();
                break;
            }
            case type::eDict:
            {
                reader.readStartDict();
                writer.writeStartDict();

                Converter::ConvertFromReader(reader, writer);

                writer.writeEndDict();
                reader.readEndDict();
                break;
            }
            case type::eTuple:
            {
                int elements = reader.readStartTuple();
                writer.writeStartTuple();

                for (int i = 0; i < elements; ++i)
                {
                    Converter::ConvertFromReader(reader, writer);
                }

                writer.writeEndTuple();
                reader.readEndTuple();
                break;
            }
            case type::eList:
            {
                reader.readStartList();
                writer.writeStartList();

                Converter::ConvertFromReader(reader, writer);

                writer.writeEndList();
                reader.readEndList();
                break;
            }
            case type::eEigenMatrix:
            {
                auto x = reader.readEigenMatrix();
                writer.writeEigenMatrix(std::get<0>(x), std::get<1>(x));
                break;
            }
            case type::eEigenQuaternion:
            {
                auto x = reader.readEigenQuaternion();
                writer.writeEigenQuaternion(x);
                break;
            }
            case type::eIVTCByteImage:
            {
                auto x = reader.readIVTCByteImage();
                writer.writeIVTCByteImage(std::get<0>(x), std::get<1>(x), std::get<2>(x));
                break;
            }
            case type::eOpenCVMat:
            {
                auto x = reader.readOpenCVMat();
                writer.writeOpenCVMat(std::get<0>(x), std::get<1>(x));
                break;
            }
            case type::ePCLPointCloud:
            {
                auto x = reader.readPCLPointCloud();
                writer.writePCLPointCloud(std::get<0>(x), std::get<1>(x), std::get<2>(x));
                break;
            }
            case type::ePosition:
            {
                reader.readPosition();
                writer.writePosition();
                break;
            }
            case type::eOrientation:
            {
                reader.readOrientation();
                writer.writeOrientation();
                break;
            }
            case type::ePose:
            {
                reader.readPose();
                writer.writePose();
                break;
            }
            case type::eInt:
            {
                reader.readInt();
                writer.writeInt();
                break;
            }
            case type::eLong:
            {
                reader.readLong();
                writer.writeLong();
                break;
            }
            case type::eFloat:
            {
                reader.readFloat();
                writer.writeFloat();
                break;
            }
            case type::eDouble:
            {
                reader.readDouble();
                writer.writeDouble();
                break;
            }
            case type::eString:
            {
                reader.readString();
                writer.writeString();
                break;
            }
            case type::eBool:
            {
                reader.readBool();
                writer.writeBool();
                break;
            }
            case type::eTime:
            {
                reader.readTime();
                writer.writeTime();
                break;
            }
            default:
            {
                throw error::DescriptorNotValidException("LegacyAronTypeReader", "ConvertFromReader", "Type could not be resolved", desc);
            }
        }
    }
}
