/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/type/primitive/Primitive.h>



namespace armarx::aron::typeIO::reader
{
    NavigatorReader::NavigatorReader(const typenavigator::NavigatorPtr& n) :
        input(n)
    {
    }

    NavigatorReader::NavigatorReader(const type::AronTypePtr& n) :
        input(typenavigator::Navigator::FromAronType(n))
    {
    }

    typenavigator::NavigatorPtr NavigatorReader::getNextAndIncrease()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        NavigatorReaderTokenPtr lastToken = stack.top();
        return lastToken->getNextElementAndIncreaseCnt();
    }

    std::tuple<std::string, int> NavigatorReader::readStartObject()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::ObjectNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::eObject, current_nav_casted);
        stack.push(newToken);
        return {newToken->getElementName(), newToken->getElementChildrenSize()};
    }

    bool NavigatorReader::readEndObject()
    {
        auto token = stack.top();
        token->assertType(type::eObject);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NavigatorReader::readStartDict()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::DictNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::eDict, current_nav_casted);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NavigatorReader::readEndDict()
    {
        auto token = stack.top();
        token->assertType(type::eDict);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NavigatorReader::readStartList()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::ListNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::eList, current_nav_casted);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NavigatorReader::readEndList()
    {
        auto token = stack.top();
        token->assertType(type::eList);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NavigatorReader::readStartTuple()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::TupleNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::eTuple, current_nav_casted);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NavigatorReader::readEndTuple()
    {
        auto token = stack.top();
        token->assertType(type::eTuple);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NavigatorReader::readStartPair()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::PairNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::ePair, current_nav_casted);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NavigatorReader::readEndPair()
    {
        auto token = stack.top();
        token->assertType(type::ePair);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    std::tuple<std::vector<int>, std::string> NavigatorReader::readEigenMatrix()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::EigenMatrixNavigator::DynamicCastAndCheck(nav);
        return std::make_tuple(casted->getDimensions(), casted->getTypename());
    }

    std::string NavigatorReader::readEigenQuaternion()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::EigenMatrixNavigator::DynamicCastAndCheck(nav);
        return casted->getTypename();
    }

    std::tuple<unsigned int, unsigned int, std::string> NavigatorReader::readIVTCByteImage()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::IVTCByteImageNavigator::DynamicCastAndCheck(nav);
        return make_tuple(casted->getWidth(), casted->getHeight(), casted->getTypename());
    }

    std::tuple<std::vector<int>, std::string> NavigatorReader::readOpenCVMat()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::OpenCVMatNavigator::DynamicCastAndCheck(nav);
        return make_tuple(casted->getDimensions(), casted->getTypename());
    }

    std::tuple<unsigned int, unsigned int, std::string> NavigatorReader::readPCLPointCloud()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::PCLPointCloudNavigator::DynamicCastAndCheck(nav);
        return make_tuple(casted->getWidth(), casted->getHeight(), casted->getTypename());
    }

    void NavigatorReader::readPosition()
    {
        auto nav = getNextAndIncrease();
        typenavigator::PositionNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readOrientation()
    {
        auto nav = getNextAndIncrease();
        typenavigator::OrientationNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readPose()
    {
        auto nav = getNextAndIncrease();
        typenavigator::PoseNavigator::DynamicCastAndCheck(nav);
    }

    // Read primitives
    void NavigatorReader::readInt()
    {
        auto nav = getNextAndIncrease();
        typenavigator::IntNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readFloat()
    {
        auto nav = getNextAndIncrease();
        typenavigator::FloatNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readLong()
    {
        auto nav = getNextAndIncrease();
        typenavigator::LongNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readDouble()
    {
        auto nav = getNextAndIncrease();
        typenavigator::DoubleNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readString()
    {
        auto nav = getNextAndIncrease();
        typenavigator::StringNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readBool()
    {
        auto nav = getNextAndIncrease();
        typenavigator::BoolNavigator::DynamicCastAndCheck(nav);
    }

    void NavigatorReader::readTime()
    {
        auto nav = getNextAndIncrease();
        typenavigator::TimeNavigator::DynamicCastAndCheck(nav);
    }

    std::string NavigatorReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NavigatorReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    // Helper functions
    type::Descriptor NavigatorReader::getTypeOfNext() const
    {
        auto token = stack.top();
        return token->getTypeOfNext();
    }
}
