/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Object.h"


namespace armarx::aron::typenavigator
{

    // constructors
    ObjectNavigator::ObjectNavigator(const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eObject, path),
        Navigator(type::Descriptor::eObject, path),
        type((new type::AronObject()))
    {
    }

    ObjectNavigator::ObjectNavigator(const std::string& name, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eObject, path),
        Navigator(type::Descriptor::eObject, path),
        type((new type::AronObject()))
    {
        setObjectName(name);
    }

    ObjectNavigator::ObjectNavigator(const type::AronObjectPtr& o, const Path& path) :
        aron::Navigator<type::Descriptor, type::AronType>::Navigator(type::Descriptor::eObject, path),
        Navigator(type::Descriptor::eObject, path),
        type(o)
    {
        CheckAronPtrForNull("ObjectNavigator", "ObjectNavigator", getPath(), o);

        for (const auto& [key, t] : o->elementTypes)
        {
            memberTypes[key] = FromAronType(t);
        }
    }

    bool ObjectNavigator::checkObjectName(const std::string& s) const
    {
        if (s.empty())
        {
            throw error::AronException("ObjectNavigator", "getResult", "The object name is empty.", getPath());
        }
        return true;
    }

    // static methods
    ObjectNavigatorPtr ObjectNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<ObjectNavigator>(n);
    }

    ObjectNavigatorPtr ObjectNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckTypeNavigatorPtrForNull("ObjectNavigator", "DynamicCast[Before]", n);
        ObjectNavigatorPtr casted = ObjectNavigator::DynamicCast(n);
        CheckTypeNavigatorPtrForNull("ObjectNavigator", "DynamicCast[After]", n->getPath(), casted);
        return casted;
    }

    // public member functions
    std::map<std::string, NavigatorPtr> ObjectNavigator::getMemberTypes() const
    {
        return memberTypes;
    }

    NavigatorPtr ObjectNavigator::getMemberType(const std::string& s) const
    {
        return memberTypes.at(s);
    }

    void ObjectNavigator::addMemberType(const std::string& k, const NavigatorPtr& v)
    {
        if (k.empty())
        {
            throw error::AronException("ObjectNavigator", "addAcceptedType", "Cannot set an element with an empty key.", getPath());
        }
        CheckTypeNavigatorPtrForNull("ObjectNavigator", "addAcceptedType", v);
        type->elementTypes[k] = v->getResult();
        memberTypes[k] = v;
    }

    void ObjectNavigator::setObjectName(const std::string& n)
    {
        //path.setRootIdentifier(n);
        type->objectName = n;
    }

    void ObjectNavigator::setExtends(const ObjectNavigatorPtr& p)
    {
        CheckTypeNavigatorPtrForNull("ObjectNavigator", "setExtends", p);
        type::AronObjectPtr ex = p->toAronObjectPtr();
        CheckAronPtrForNull("ObjectNavigator", "setExtends", getPath(), ex);
        extends = p;
    }

    std::string ObjectNavigator::getObjectName() const
    {
        return type->objectName;
    }

    ObjectNavigatorPtr ObjectNavigator::getExtends() const
    {
        return extends;
    }

    std::vector<std::string> ObjectNavigator::getAllKeys() const
    {
        std::vector<std::string> ret;
        for (const auto& [k, _] : memberTypes)
        {
            ret.push_back(k);
        }
        return ret;
    }

    type::AronObjectPtr ObjectNavigator::toAronObjectPtr() const
    {
        checkObjectName(type->objectName);

        // TODO: Shall we allow empty objects?
        //if(acceptedTypeNavigators.empty())
        //{
        //    throw exception::AronExceptionWithPathInfo("ObjectNavigator", "getResult", "No accepted types set", getPath());
        //}
        return type;
    }

    // virtual implementations
    std::vector<NavigatorPtr> ObjectNavigator::getChildren() const
    {
        std::vector<NavigatorPtr> ret;
        for (const auto& [k, t] : memberTypes)
        {
            ret.push_back(t);
        }
        return ret;
    }

    size_t ObjectNavigator::childrenSize() const
    {
        return memberTypes.size();
    }

    type::AronTypePtr ObjectNavigator::getResult() const
    {
        return toAronObjectPtr();
    }

    std::string ObjectNavigator::getName() const
    {
        return "AronObjectType<" + type->objectName + ">";
    }
}

